package livecld;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedHashMap;
import java.util.LinkedList;

import livecld.model.HierarchyNodeModel;
import livecld.model.MessageModel;

import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.jdt.core.Flags;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IMember;
import org.eclipse.jdt.core.IPackageFragment;
import org.eclipse.jdt.core.IPackageFragmentRoot;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.ITypeHierarchy;
import org.eclipse.jdt.core.JavaModelException;


public class HierarchyModelProvider {
	
	public static final int PUBLIC = 4;
	public static final int DEFAULT = 3;
	public static final int PROTECTED = 2;
	public static final int PRIVATE = 1;
	
	
	
	
	
	public static int getVisibility(IMember element) {
		try {
			int f = element.getFlags();
			
			if (Flags.isPublic(f))
				return PUBLIC;
			else if (Flags.isProtected(f))
				return PROTECTED;
			else if (Flags.isPrivate(f))
				return PRIVATE;
			else
				return DEFAULT;
		} catch (JavaModelException e) {
			e.printStackTrace();
		}
		
		return -1;
	}
	
	
	
	private IJavaProject projectFilter = null;
	private IPackageFragment packageFilter = null;
	private int visibility = PRIVATE;
	
	public void setProjectFilter(IJavaProject projectFilter) {
		this.projectFilter = projectFilter;
	}
	
	public void setPackageFilter(IPackageFragment packageFilter) {
		this.packageFilter = packageFilter;
	}
	
	public void setVisibility(int visibility) {
		this.visibility = visibility;
	}
	
	
	public Object getTypes() {
		
		LinkedList<IType> types = new LinkedList<IType>();
		LinkedHashMap<IType, HierarchyNodeModel> map = new LinkedHashMap<IType, HierarchyNodeModel>();
		
		try {
			ArrayList<IPackageFragment> packages = new ArrayList<IPackageFragment>();
			
			if (packageFilter == null) {
				if (projectFilter != null) {
					//use project filter
					for (IPackageFragmentRoot packageFragmentRoot : projectFilter.getPackageFragmentRoots()) {
						if (!packageFragmentRoot.isArchive())
							for (IJavaElement packageFragment : packageFragmentRoot.getChildren()) {
								packages.add((IPackageFragment) packageFragment);
							}
					}
				}
			} else {
				//use package filter
				packages.add(packageFilter);
			}

			
			for (IJavaElement packageFragment : packages) {
				for (ICompilationUnit compilationUnit : ((IPackageFragment)packageFragment).getCompilationUnits())
					for (IType leafType : compilationUnit.getAllTypes()) {
						
						if (getVisibility(leafType) >= visibility)
							types.add(leafType);
					}
			}

			
			
			Collections.sort(types, new JavaElementComparator<IType>());
			
			for (IType type : types) {
				if (map.containsKey(type))
					continue;
				
				ITypeHierarchy hierarchy = type.newSupertypeHierarchy(new NullProgressMonitor());

				HierarchyNodeModel child = new HierarchyNodeModel(type);
				map.put(type, child);
				
				for (IType t : hierarchy.getAllSuperclasses(type)) {
					if (map.containsKey(t)) {
						HierarchyNodeModel parent = map.get(t);
						parent.addChild(child);
						break;
					} else {
						HierarchyNodeModel parent = new HierarchyNodeModel(t);
						map.put(t, parent);
						
						parent.addChild(child);
						child = parent;
					}
				}
			}
		} catch (JavaModelException e) {
			e.printStackTrace();
		}
		
		
		
		
		if (map.values().size() < 1) {
			return new MessageModel("Nothing to show.");
		}
		
		HierarchyNodeModel root = map.values().iterator().next();
		
		while (root.getParent() != null)
			root = root.getParent();
		
		return root;
	}
	
	
	private static class JavaElementComparator<T> implements Comparator<T> {
		public int compare(T a, T b) {
			if (a instanceof IType && b instanceof IType) {
				IType ea = (IType) a;
				IType eb = (IType) b;
				
				return ea.getFullyQualifiedName().compareTo(eb.getFullyQualifiedName());
			}
			
			return a.hashCode() - b.hashCode();
		}
	}
	
	
}
