package livecld;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Iterator;

import livecld.controller.ClassDiagrammPartFactory;
import livecld.controller.DiagrammRootEditPart;
import livecld.javamodel.JavaModelListener;
import livecld.javamodel.filter.JavaElementTypeFilter;
import livecld.model.DiagrammRootModel;

import org.eclipse.draw2d.ColorConstants;
import org.eclipse.draw2d.Graphics;
import org.eclipse.draw2d.IFigure;
import org.eclipse.draw2d.SWTGraphics;
import org.eclipse.draw2d.geometry.Rectangle;
import org.eclipse.gef.ContextMenuProvider;
import org.eclipse.gef.EditDomain;
import org.eclipse.gef.EditPart;
import org.eclipse.gef.GraphicalViewer;
import org.eclipse.gef.LayerConstants;
import org.eclipse.gef.MouseWheelHandler;
import org.eclipse.gef.MouseWheelZoomHandler;
import org.eclipse.gef.editparts.ScalableFreeformRootEditPart;
import org.eclipse.gef.ui.parts.GraphicalViewerKeyHandler;
import org.eclipse.gef.ui.parts.ScrollingGraphicalViewer;
import org.eclipse.jdt.core.ElementChangedEvent;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IElementChangedListener;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IJavaElementDelta;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IMember;
import org.eclipse.jdt.core.IPackageFragment;
import org.eclipse.jdt.core.IType;
import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.IToolBarManager;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Device;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.graphics.ImageLoader;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.FileDialog;
import org.eclipse.ui.IActionBars;
import org.eclipse.ui.ISelectionListener;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.actions.ActionGroup;
import org.eclipse.ui.part.ViewPart;

import resources.ResourceManager;

public class HierarchyViewPart extends ViewPart {
	private ScrollingGraphicalViewer viewer;

	private EditDomain editDomain;

	private DiagrammRootModel modelRoot;

	private boolean currentProjectFilter = true;

	private boolean currentPackageFilter = false;

	private boolean showBinaryMembers = false;

	private int visibility = HierarchyModelProvider.PRIVATE;

	private IJavaProject currentProject;

	private IPackageFragment currentPackage;

	/**
	 * The constructor.
	 */
	public HierarchyViewPart() {
		modelRoot = new DiagrammRootModel();
	}

	/**
	 * This is a callback that will allow us to create the viewer and initialize
	 * it.
	 */
	@Override
	public void createPartControl(Composite parent) {
		editDomain = new ViewPartEditDomain(this);

		viewer = new ScrollingGraphicalViewer();
		viewer.createControl(parent);
		viewer.getControl().setBackground(ColorConstants.listBackground);

		editDomain.addViewer(viewer);

		viewer.setEditPartFactory(new ClassDiagrammPartFactory());

		ScalableFreeformRootEditPart root = new ScalableFreeformRootEditPart();

		double zoomlevels[] = { 0.025, 0.05, 0.1, 0.25, 0.5, 0.75, 1, 1.5, 2 };

		root.getZoomManager().setZoomLevels(zoomlevels);
		root.getZoomManager().setZoom(1);

		viewer.setRootEditPart(root);

		viewer.setKeyHandler(new GraphicalViewerKeyHandler(viewer));
		viewer.setProperty(MouseWheelHandler.KeyGenerator.getKey(SWT.MOD1), MouseWheelZoomHandler.SINGLETON);

		viewer.setContents(modelRoot);

		contributeToActionBars();

		ViewPart viewPart = this;
		viewPart.getSite().getWorkbenchWindow().getSelectionService()
							.addSelectionListener(selectionListener);

		// configure the context menu provider
		ContextMenuProvider cmProvider = new HierarchyViewContextMenuProvider(this);
		viewer.setContextMenu(cmProvider);
		getSite().registerContextMenu(cmProvider, viewer);

		JavaModelListener.getListener().addChangedListener(new ReloadOnAddOrRemoveChanges(), new JavaElementTypeFilter(ICompilationUnit.class));
		JavaModelListener.getListener().addChangedListener(new ReloadOnAddOrRemoveChanges(), new JavaElementTypeFilter(IType.class));

		refresh();
	}

	public GraphicalViewer getGraphicalViewer() {
		return viewer;
	}

	public EditDomain getEditDomain() {
		return editDomain;
	}

	public void setFocus() {

	}

	public void refresh() {
		HierarchyModelProvider p = new HierarchyModelProvider();

		if (currentPackageFilter) {
			p.setPackageFilter(currentPackage);
			p.setProjectFilter(currentProject);
		} else if (currentProjectFilter) {
			p.setProjectFilter(currentProject);
		}

		p.setVisibility(visibility);

		modelRoot.setChildren(p.getTypes());

		DiagrammRootEditPart diagrammRootEditPart = (DiagrammRootEditPart) viewer.getContents();
		diagrammRootEditPart.setVisibility(visibility);
		diagrammRootEditPart.setShowBinaryMembers(showBinaryMembers);

		new EditPartRefreshJob(diagrammRootEditPart, true).schedule();
	}

	private class RefreshAction extends Action {
		public RefreshAction() {
			setText("Refresh");
			setToolTipText("Refresh Class Diagram");
			setImageDescriptor(ResourceManager.getImageDescriptor("refresh.gif"));
		}

		@Override
		public void run() {
			refresh();
		}
	}

	private void contributeToActionBars() {
		IActionBars bars = getViewSite().getActionBars();
		fillLocalPullDown(bars.getMenuManager());
		fillLocalToolBar(bars.getToolBarManager());
	}

	private void fillLocalPullDown(IMenuManager manager) {

	}

	private void fillLocalToolBar(IToolBarManager manager) {
		manager.add(new RefreshAction());
	}

	private ISelectionListener selectionListener = new ISelectionListener() {
		public void selectionChanged(IWorkbenchPart sourcepart, ISelection selection) {
			// we ignore our own selections
			if (sourcepart != HierarchyViewPart.this) {
				if (selection instanceof IStructuredSelection) {
					Iterator it = ((IStructuredSelection) selection).iterator();

					Object last = null;

					while (it.hasNext()) {
						last = it.next();
					}

					if (last == null)
						return;

					boolean reload = false;

					EditPart part = (EditPart) viewer.getEditPartRegistry().get(last);

					if (part != null) {
						viewer.reveal(part);
					} else if (last instanceof IJavaElement) {
						IJavaProject project = null;
						IPackageFragment packageFragment = null;

						if (last instanceof IJavaProject) {
							project = (IJavaProject) last;
							packageFragment = null;
						} else if (last instanceof IPackageFragment) {
							packageFragment = (IPackageFragment) last;
						} else if (last instanceof ICompilationUnit) {
							ICompilationUnit cu = (ICompilationUnit) last;
							packageFragment = cu.findPrimaryType().getPackageFragment();
						} else if (last instanceof IType) {
							IType type = (IType) last;
							packageFragment = type.getPackageFragment();
						} else if (last instanceof IMember) {
							IMember member = (IMember) last;
							IType type = member.getDeclaringType();

							if (type != null) {
								packageFragment = type.getPackageFragment();
							}
						}

						if (packageFragment != null)
							project = packageFragment.getJavaProject();

						if (project == null)
							return;

						if (currentProject != project || currentPackage != packageFragment) {
							reload = true;
							currentProject = project;
							currentPackage = packageFragment;
						}
					}

					if (reload)
						refresh();
				}
			}
		}
	};

	public void dispose() {
		getSite().getWorkbenchWindow().getSelectionService().removeSelectionListener(selectionListener);
		super.dispose();
	}

	public ActionGroup getStandardActionGroup() {
		return standardActionGroup;
	}

	private class ReloadOnAddOrRemoveChanges implements IElementChangedListener {
		public void elementChanged(ElementChangedEvent evt) {
			IJavaElementDelta delta = evt.getDelta();

			if (delta.getKind() == IJavaElementDelta.ADDED || delta.getKind() == IJavaElementDelta.REMOVED) {
				refresh();
			} else if (delta.getKind() == IJavaElementDelta.CHANGED) {
				if ((delta.getFlags() & (IJavaElementDelta.F_CONTENT | IJavaElementDelta.F_MODIFIERS)) != 0) {
					refresh();
				}
			}
		}
	}

	private ActionGroup standardActionGroup = new ActionGroup() {
		@Override
		public void fillContextMenu(IMenuManager manager) {
			MenuManager vis = new MenuManager("Visibility");
			vis.add(new SetVisibilityAction(HierarchyModelProvider.PUBLIC));
			vis.add(new SetVisibilityAction(HierarchyModelProvider.DEFAULT));
			vis.add(new SetVisibilityAction(HierarchyModelProvider.PROTECTED));
			vis.add(new SetVisibilityAction(HierarchyModelProvider.PRIVATE));
			manager.add(vis);

			MenuManager fil = new MenuManager("Filter");
			fil.add(new ToggleBinaryMembersAction());
			fil.add(new ToggleCurrentProjectFilterAction());
			fil.add(new ToggleCurrentPackageFilterAction());
			manager.add(fil);
			
			manager.add(new ExportRasterAction());
		}
	};

	private class SetVisibilityAction extends Action {
		private int setv;

		public SetVisibilityAction(int setv) {
			this.setv = setv;

			switch (setv) {
			case HierarchyModelProvider.PUBLIC:
				setText("public");
				break;
			case HierarchyModelProvider.PROTECTED:
				setText("protected");
				break;
			case HierarchyModelProvider.PRIVATE:
				setText("private");
				break;
			case HierarchyModelProvider.DEFAULT:
				setText("default");
				break;
			}

			setChecked(visibility == setv);
		}

		@Override
		public void run() {
			visibility = setv;
			refresh();
		}
	}

	private class ToggleBinaryMembersAction extends Action {
		public ToggleBinaryMembersAction() {
			setText("Members of Binary Types");
			setChecked(showBinaryMembers);
		}

		@Override
		public void run() {
			showBinaryMembers = !showBinaryMembers;
			refresh();
		}
	}

	private class ToggleCurrentProjectFilterAction extends Action {
		public ToggleCurrentProjectFilterAction() {
			setText("Currently Selected Project");
			setChecked(currentProjectFilter);
		}

		@Override
		public void run() {
			currentProjectFilter = !currentProjectFilter;

			if (!currentProjectFilter)
				currentPackageFilter = false;

			refresh();
		}
	}

	private class ToggleCurrentPackageFilterAction extends Action {
		public ToggleCurrentPackageFilterAction() {
			setText("Currently Selected Package");
			setChecked(currentPackageFilter);
		}

		@Override
		public void run() {
			currentPackageFilter = !currentPackageFilter;

			if (currentPackageFilter)
				currentProjectFilter = true;

			refresh();
		}
	}

	private class ExportRasterAction extends Action {
		public ExportRasterAction() {
			setText("Export");
		}

		@Override
		public void run() {
			FileDialog fileChooser = new FileDialog(getSite().getShell(), SWT.SAVE);
			fileChooser.setText("Export");
			fileChooser.setFileName("diagram.jpeg");

			fileChooser.setFilterExtensions(new String[] { "*.jpeg", "*.bmp" });
			fileChooser.setFilterNames(new String[] { "JPEG", "BMP" });

			String filename = fileChooser.open();
			if (filename != null) {
				ScalableFreeformRootEditPart root = (ScalableFreeformRootEditPart) viewer.getRootEditPart();
				IFigure print = root.getLayer(LayerConstants.SCALABLE_LAYERS);

				try {
					createImage(print, new File(filename));
				} catch (FileNotFoundException e) {
					e.printStackTrace();
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		}

		private void createImage(IFigure figure, File file) throws IOException {
			Device device = viewer.getControl().getDisplay();
			Rectangle r = figure.getBounds();

			FileOutputStream out = new FileOutputStream(file);

			Image image = null;
			GC gc = null;
			Graphics g = null;
			try {
				image = new Image(device, r.width, r.height);
				gc = new GC(image);
				g = new SWTGraphics(gc);
				g.translate(r.x * -1, r.y * -1);

				figure.paint(g);
				
				int format = SWT.IMAGE_JPEG;
				
				if (file.getName().endsWith(".bmp")) {
					format = SWT.IMAGE_BMP;
				}
				
				//exporting GIF and PNG doesn't work for some reason

				ImageLoader imageLoader = new ImageLoader();
				imageLoader.data = new ImageData[] { image.getImageData() };
				imageLoader.save(file.getAbsolutePath(), format);
			} finally {
				if (g != null) {
					g.dispose();
				}
				if (gc != null) {
					gc.dispose();
				}
				if (image != null) {
					image.dispose();
				}
				if (out != null) {
					out.close();
				}
			}
		}

	}

}
